-------------------------------------------------------------------------------
-- Localization libraries
-------------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("SimpleSelfRebuff")
local BS = AceLibrary("Babble-Spell-2.2")

-------------------------------------------------------------------------------
-- Aliases & constants
-------------------------------------------------------------------------------

local activationCheckEvents = {
	'PLAYER_DEAD', 'PLAYER_UNGHOST', 'PLAYER_ALIVE',
	'PLAYER_UPDATE_RESTING',
	'PLAYER_CONTROL_GAINED', 'PLAYER_CONTROL_LOST',
	'SimpleSelfRebuff_CheckActivation'
}

local CATEGORY_TRACKING = L["Tracking"] or "Tracking"
local CATEGORY_MAINHAND = L["Main weapon"] or "Main weapon"
local CATEGORY_OFFHAND  = L["Off-hand weapon"] or "Off-hand weapon"

local SOURCE_SPELL = 'spell'
local SOURCE_ITEM  = 'item'

local TARGET_AURA     = 'aura'
local TARGET_TRACKING = 'tracking'
local TARGET_MAINHAND = 'main-hand'
local TARGET_OFFHAND  = 'off-hand'

-------------------------------------------------------------------------------
-- Locals
-------------------------------------------------------------------------------

local categories = {}

local buffToCast
local castButton

local monitoringActive

local db

local sources = {}
local targets = {}

local loadFuncs = {}

-------------------------------------------------------------------------------
-- Externals
-------------------------------------------------------------------------------

local AceOO = AceLibrary('AceOO-2.0')
local Gratuity = AceLibrary('Gratuity-2.0')
local Waterfall = AceLibrary:HasInstance('Waterfall-1.0') and AceLibrary('Waterfall-1.0')
 
local Mixin = AceOO.Mixin
local Class = AceOO.Class
local Classpool = AceOO.Classpool

-------------------------------------------------------------------------------
-- Addon declaration
-------------------------------------------------------------------------------

SimpleSelfRebuff = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "AceModuleCore-2.0")
local SimpleSelfRebuff = SimpleSelfRebuff

-- Static early initialization
do
	local self = SimpleSelfRebuff

	self:RegisterDB("SimpleSelfRebuffDB")

	self.options = {	
		type = 'group',
		args = {
			menu = {
				type = 'execute',
				name = L["Menu"],
				desc = L["Menu"],
				func = function() Waterfall:Open('SimpleSelfRebuff') end,
				hidden = not Waterfall,
				wfHidden = true,
				order = 50,
			},
			menuHeader = {
				type = 'header',
				name = L["Global options"],
				order = 51,
				wfHidden = true,					
				hidden = not Waterfall,
			},		
			--
			rebuffThreshold = {
				type = 'range',
				name = L["Rebuff time threshold"],
				desc = L["Any buff with remaining time less than this value will be recast. Select 0 to disable."],
				min = 0,
				max = 180,
				step = 1,
				bigStep = 5,
				get = function() return db.rebuffThreshold end,
				set = function(value)
					db.rebuffThreshold = value
					self:CheckRebuff()
				end,
			},
			--
			buffHeader = {
				type = 'header',
				name = L["Buff selection"],
				order = 199,
				wfHidden = true,
			},
		}
	}
		
	self.CATEGORY_TRACKING = CATEGORY_TRACKING
	self.CATEGORY_MAINHAND = CATEGORY_MAINHAND 
	self.CATEGORY_OFFHAND  = CATEGORY_OFFHAND 
	
	self.SOURCE_SPELL = SOURCE_SPELL
	self.SOURCE_ITEM  = SOURCE_ITEM
	
	self.TARGET_AURA     = TARGET_AURA
	self.TARGET_TRACKING = TARGET_TRACKING 
	self.TARGET_MAINHAND = TARGET_MAINHAND
	self.TARGET_OFFHAND  = TARGET_OFFHAND

	-- For debugging
	SSR = self
	
	self.categories = categories
	self.sources = sources
	self.targets = targets
end

-------------------------------------------------------------------------------
-- Helpers
-------------------------------------------------------------------------------

local function error(msg, ...)
	if select('#', ...) > 0 then
		msg = msg:format(...)
	end
	return _G.error(msg, 2)
end

local function erase(...)
	for i=1,select('#', ...) do
		local t = select(i, ...)
		if type(t) == 'table' then
			for k in pairs(t) do
				t[k] = nil 
			end
		end
	end
end

local function loadHash(t, ...)
	for i = 1, select('#', ...), 2 do
		local k,v = select(i, ...)
		if t[k] == nil then
			t[k] = v
		end
	end
end

local function colorize(text, r, g, b)
	return ("|cff%02x%02x%02x%s|r"):format(math.floor(r*255), math.floor(g*255), math.floor(b*255), text)
end
SimpleSelfRebuff.colorize = colorize

local new, del
do
	local list = setmetatable({}, {__mode='k'})
	
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return {...}
		end
	end
	function del(t)
		if type(t) == 'table' then
			setmetatable(t, nil)
			for k in pairs(t) do
				t[k] = nil
			end
			list[t] = true
		end
		return nil
	end
end

local function formatDuration(duration, default)
	if not duration or duration < 1 then
		return default or "-"
	elseif duration >= 3600 then
		return HOUR_ONELETTER_ABBR:format(math.ceil(duration/3600))
	elseif duration >= 60 then
		return MINUTE_ONELETTER_ABBR:format(math.ceil(duration/60))
	else
		return SECOND_ONELETTER_ABBR:format(duration)
	end
end
SimpleSelfRebuff.formatDuration = formatDuration

-------------------------------------------------------------------------------
-- Initializing
-------------------------------------------------------------------------------

function SimpleSelfRebuff:_rev(tag)
	self.revision = math.max(self.revision or 0, tonumber(tag:match("%d+")))
end

-- Addon setup
function SimpleSelfRebuff:OnInitialize()
	self:_rev("$Revision: 43066 $")
	self.version = "r" .. SimpleSelfRebuff.revision
	self.date = string.sub("$Date: 2007-07-07 06:57:29 -0400 (Sat, 07 Jul 2007) $", 8, 17)

	self:RegisterDefaults('profile', {
		categories = {},
		rebuffThreshold = 30,
	})
		
	db = self.db.profile

	self:RegisterChatCommand({"/ssr", "/simpleselfrebuff"}, self.options)
	
	local AceDebug = AceLibrary:HasInstance('AceDebug-2.0') and AceLibrary('AceDebug-2.0')
	if AceDebug then
		AceDebug:embed(self)
		local opt = AceDebug:GetAceOptionsDataTable(self)
		self.options.args.debug = opt.debug
		--SimpleSelfRebuff:SetDebugging(UnitName('player') == 'Adirelle')
	else
		function SimpleSelfRebuff.Debug() end
	end
	
	if Waterfall then
		Waterfall:Register('SimpleSelfRebuff', 
			'aceOptions', self.options,
			'title','SimpleSelfRebuff',
			'colorR',0.8,
			'colorG',0.2,
			'colorB',0.8
		)
	end
		
end

function SimpleSelfRebuff:OnEnable(first)
	if first then
		self:ProcessRegistry()
	end

	self:RegisterBucketEvent(activationCheckEvents, 0.5, 'CheckActivation')
	self:RegisterBucketEvent('PLAYER_AURAS_CHANGED', 0)

	self:CheckActivation()
end

function SimpleSelfRebuff:OnDisable(first)
	self:SetMonitoringActive(false)
end

function SimpleSelfRebuff:PLAYER_AURAS_CHANGED()
	self:CheckActivation()	
	self:ScanPlayerBuffs()
end

function SimpleSelfRebuff:AvailableBuffsChanged()
	self:UpdateOptions()
	self:CheckActivation()
end

-------------------------------------------------------------------------------
-- Module modulePrototype
-------------------------------------------------------------------------------

SimpleSelfRebuff:SetModuleMixins('AceEvent-2.0')
SimpleSelfRebuff.modulePrototype.core = SimpleSelfRebuff

function SimpleSelfRebuff:OnModuleCreate(name, module)
	self:SetModuleDefaultState(module, true)
end

function SimpleSelfRebuff.modulePrototype:OnInitialize()
	self.db = self.core:AcquireDBNamespace(self.name)
	
	if type(self.OnModuleInitialize) == "function" then
		self:OnModuleInitialize()
	end
	
	local options = self.options or {}
	
	options.enable = {
		type = 'toggle',
		name = L['Enable'],
		desc = L['Enable %s.'];format(self.optionName or self.name),
		order = 1,
		handler = self.core,
		passValue = self,
		get = "IsModuleActive",
		set = "ToggleModuleActive",
	}
	
	self.core.options.args[self.name] = {
		type = 'group',
		name = self.optionName or self.name,
		desc = self.optionDsc or self.name,
		handler = self,
		args = options,
		order = self.optionOrder,
	}	
	
end

function SimpleSelfRebuff.modulePrototype:OnEnable(first)
	if type(self.OnModuleEnable) == "function" then
		self:OnModuleEnable(first)
	end
end

function SimpleSelfRebuff.modulePrototype:OnDisable()
	if type(self.OnModuleDisable) == "function" then
		self:OnModuleDisable()
	end
end

function SimpleSelfRebuff.modulePrototype:RegisterDefaults(kind, defaults)
	self.core:RegisterDefaults(self.name, kind, defaults)
end

-------------------------------------------------------------------------------
-- Options handling
-------------------------------------------------------------------------------

do
	local NO_BUFF = L["None"]
	
	local function getSelectCategoryBuff(category)
		local buff = category:GetExpectedBuff()
		return buff and buff.name or NO_BUFF
	end
	
	local function getToggleCategoryBuff(category)
		local buff = category:GetExpectedBuff()
		return buff and (buff.name == category.name)
	end
	
	local function setCategoryBuff(category, name)
		if name == NO_BUFF then
			name = nil
		elseif name == true then
			name = category.name
		end
		category:SetExpectedBuff(name)
	end
	
	function SimpleSelfRebuff:BuildOptions()
		local opts = self.options.args
		for categoryName, category in pairs(categories) do
			local categoryName, category = categoryName, category
			local option = {
				name = categoryName,
				desc = categoryName,
				passValue = category,
				set = setCategoryBuff,
				order = 200,
			}
			if category.count > 1 then
				option.type = 'text'
				option.get = getSelectCategoryBuff
				local validate = { NO_BUFF }
				option.validate = validate
				option.disabled = function() return #validate < 2 end
			else
				option.type = 'toggle'
				option.get = getToggleCategoryBuff
				local theBuff = select(2, next(category.buffs))
				option.disabled = function() return not theBuff:IsCastable() end
			end
			opts[categoryName] = option
		end
		self:UpdateOptions()
	end
	
	function SimpleSelfRebuff:UpdateOptions()
		for categoryName, category in pairs(categories) do
			if category.count > 1 then
				local validate = self.options.args[category.name].validate
				erase(validate)
				for name,buff in pairs(category.buffs) do
					if buff:IsCastable() then
						tinsert(validate, buff.name)
					end
				end
				table.sort(validate)				
				tinsert(validate, 1, NO_BUFF)
			end
		end
	end
end

-------------------------------------------------------------------------------
-- Base buff class
-------------------------------------------------------------------------------
local CategoryClass, BuffClass

BuffClass = Class()

function BuffClass.prototype:init(name, category, ...)
	if type(name) ~= 'string' then
		error('Argument #2 to BuffClass:new() should be a string, not %q', type(name))
	end		
	
	self.name = name
	self.category = category
	loadHash(self, ...)
	
	BuffClass.super.prototype.init(self)
	
	local targetMixin = targets[self.target]
	targetMixin.buffs[self.name] = self
	targetMixin.allBuffs[self] = true
	
	local sourceMixin = sources[self.source]
	sourceMixin.buffs[self.name] = self
	sourceMixin.allBuffs[self] = true
end

function BuffClass.prototype:IsCastable()
	if self.found then
		if self.minLevel and UnitLevel('player') < self.minLevel then
			return false
		elseif self.maxLevel and UnitLevel('player') > self.maxLevel then
			return false
		elseif type(self.IsTargetCastable) == "function" then
			return self:IsTargetCastable()
		end
		return true
	end
	return false
end

-------------------------------------------------------------------------------
-- Category class
-------------------------------------------------------------------------------

CategoryClass = Class()

function CategoryClass.prototype:init(name, target, source, ...)
	if type(name) ~= 'string' then
		error('Argument #2 to CategoryClass:new() should be a string, not %q', type(name))
	end
	CategoryClass.super.prototype.init(self)
	self.name = name
	self.buffs = {}
	self.count = 0
	self.target = target or TARGET_AURA
	self.source = source or SOURCE_SPELL
	loadHash(self, ...)
end

function CategoryClass.prototype:add(name, ...)

	if self.buffs[name] then
		error("%q already registered in category %q", name, self.name)
	end
	
	local source = self.source
	local target = self.target
	
	if not sources[source] and loadFuncs[source] then
		loadFuncs[source]()
		loadFuncs[source] = nil
	end
	if not sources[source] then
		error("unknown source for %q: %q", name, source)
	end		
	
	if not targets[target] and loadFuncs[target] then
		loadFuncs[target]()
		loadFuncs[target] = nil
	end	
	if not targets[target] then
		error("unknown target for %q: %q", name, target)
	end		
		
	local class = Classpool(BuffClass, sources[source], targets[target])
	local buff = class:new(name, self, ...)

	self.buffs[buff.name] = buff
	self.count = self.count + 1

	return self
end

function CategoryClass.prototype:addMulti(...)
	for i=1,select('#', ...) do
		local name = select(i, ...)
		self:add(name)
	end
	return self
end

function CategoryClass.prototype:GetExpectedBuff()
	local name = db.categories[self.name]
	if self.buffs[name] then
		return self.buffs[name]
	else
		db.categories[self.name] = nil
	end
end

function CategoryClass.prototype:SetExpectedBuff(buff)
	local buffName 
	if type(buff) == 'string' then
		buffName = buff
	elseif buff then
		buffName = buff.name
	end
	if buffName and not self.buffs[buffName] then
		error("Unknown buff %q in category %q", buffName, self.name)
	end
	db.categories[self.name] = buffName
	SimpleSelfRebuff:ScanPlayerBuffs()
	SimpleSelfRebuff:CheckRebuff()
end

function CategoryClass.prototype:GetActualBuff()
	local timeLeft = self.actualBuffExpiration and math.floor(self.actualBuffExpiration - GetTime())
	if not timeLeft or timeLeft <= 0 then
		timeLeft = nil
	end
	return self.actualBuff, self.actualBuffExpiration, timeLeft
end

function CategoryClass.prototype:SetActualBuff(buff, expiration)
	if not buff or not expiration or expiration == 0 then
		expiration = nil
	end
	local dirty = buff ~= self.actualBuff or expiration ~= self.actualBuffExpiration
	self.actualBuff = buff
	self.actualBuffExpiration = expiration
	return dirty
end

-------------------------------------------------------------------------------
-- Some mixin helpers
-------------------------------------------------------------------------------

local BuffMixin, BuffSourceMixin, BuffTargetMixin

do
	
	local function returnTrue() return true end

	function BuffMixin(...)
		local mixin = Mixin{...}
		mixin.allBuffs = {}
		mixin.buffs = {}
		return mixin
	end
	
	function BuffSourceMixin(source, ...)
		mixin = BuffMixin("IsUsable", "SetupBuffSource", "source", ...)
		mixin.source = source
		sources[source] = mixin
		return mixin
	end
	
	function BuffTargetMixin(target, ...)
		mixin = BuffMixin("IsTargetCastable", "SetupBuffTarget", "target", ...)
		mixin.target = target
		mixin.IsTargetCastable = returnTrue
		targets[target] = mixin
		return mixin
	end
	
end

-------------------------------------------------------------------------------
-- Buffs from items
-------------------------------------------------------------------------------

loadFuncs[SOURCE_ITEM] = function()
	SimpleSelfRebuff:Debug("Loading buff-from-item support")
	
	local ItemBuff = BuffSourceMixin(SOURCE_ITEM)
	local missingItems
	
	local function scheduleScan()
		SimpleSelfRebuff:ScheduleEvent('SimpleSelfRebuff_ScanItems', 0.5, ItemBuff.ScanItems, ItemBuff)
	end
	
	local function enableItemScan(flag)
		if flag and not SimpleSelfRebuff:IsEventRegistered('BAG_UPDATE') then
			SimpleSelfRebuff:RegisterEvent('BAG_UPDATE', scheduleScan)
		elseif SimpleSelfRebuff:IsEventRegistered('BAG_UPDATE') then
			SimpleSelfRebuff:UnregisterEvent('BAG_UPDATE')
		end
	end
	
	function ItemBuff:OnInstanceInit(buff)
		if not buff.itemid then
			error("itemid is missing for ItemBuff %q", buff.name)
		end
		if not missingItems then
			missingItems = new()
		end
		missingItems[buff.itemid] = true
	end
	
	function ItemBuff:OnEnable()
		if next(self.allBuffs) and missingItems then
			self:ScanItems()
		end
	end
	
	function ItemBuff:OnDisable()
		enableItemScan(false)
	end
	
	do
		local itemsToCache
		
		local function CacheItem()
			local itemid = itemsToCache and next(itemsToCache)
			if itemid then
				itemsToCache[itemid] = nil
				SimpleSelfRebuff:Debug('Caching %d', itemid)	
				Gratuity:SetHyperlink("item:"..itemid)
			end
			if not next(itemsToCache) then
				SimpleSelfRebuff:CancelScheduledEvent('SimpleSelfRebuff_CacheItems')
				itemsToCache = del(itemsToCache)
				SimpleSelfRebuff:Debug('Caching done, rescanning')	
				ItemBuff:ScanItems()
			end
		end
	
		function ItemBuff:ScanItems()
			if not missingItems then
				enableItemScan(false)
				return
			end
		
			local dirty
		
			for buff in pairs(self.allBuffs) do
				if not buff.itemid then
					error("Missing itemid for %q", buff.name)
				end
				local itemName = GetItemInfo(buff.itemid)
				if not itemName then
					if not itemsToCache then
						itemsToCache = new()
					end
					itemsToCache[buff.itemid] = true
				else
					if not buff.found then
						buff.found = true							
						dirty = true			
					end
					if buff.name ~= itemName then
						buff.name = itemName
						dirty = true			
					end
					if missingItems then
						missingItems[buff.itemid] = nil
					end
				end
			end
			
			if itemsToCache then
				SimpleSelfRebuff:Debug('Starting caching')	
				SimpleSelfRebuff:ScheduleRepeatingEvent('SimpleSelfRebuff_CacheItems', CacheItem, 0.4)
				enableItemScan(false)
			else
				if missingItems and not next(missingItems) then
					missingItems = del(missingItems)
				end
				enableItemScan(missingItems and next(missingItems))					
			end
			
			if dirty then
				for categoryName, category in pairs(categories) do
					local buffs = category.buffs
					for buffName, buff in pairs(buffs) do
						if buffName ~= buff.name then
							SimpleSelfRebuff:Debug('%q => %q', buffName, buff.name)
							buffs[buff.name] = nil
							buffs[buffName] = buff
						end
					end
				end
				
				SimpleSelfRebuff:AvailableBuffsChanged()
			end
		end
	end
	
	function ItemBuff:IsUsable()
		local start, duration = GetItemCooldown(self.itemid)
		if start > 0 and duration > 0 and start + duration > GetTime() then
			return false
		end					
		return (IsUsableItem(self.itemid))
	end
	
	function ItemBuff:SetupBuffSource(castButton)
		SimpleSelfRebuff:Debug('Setup using for item %q', self.name)
		castButton:SetAttribute('type', 'item')
		castButton:SetAttribute('item', self.name)
	end
	
end

-------------------------------------------------------------------------------
-- Buffs from spells
-------------------------------------------------------------------------------

loadFuncs[SOURCE_SPELL] = function()
	SimpleSelfRebuff:Debug('Loading buff-from-spell support')
		
	local SpellBuff = BuffSourceMixin(SOURCE_SPELL)
	
	local function scheduleScan()
		SimpleSelfRebuff:ScheduleEvent('SimpleSelfRebuff_ScanSpells', 0.5, SpellBuff.ScanSpells, SpellBuff)
	end	
	
	function SpellBuff:OnEnable()
		if next(self.allBuffs) then
			SimpleSelfRebuff:RegisterEvent('SPELLS_CHANGED', scheduleScan)
			self:ScanSpells()
		end
	end
	
	function SpellBuff:OnDisable()
		if SimpleSelfRebuff:IsEventRegistered('SPELLS_CHANGED') then
			SimpleSelfRebuff:UnregisterEvent('SPELLS_CHANGED')
		end
	end
	
	function SpellBuff:ScanSpells()
		local dirty = false
		local knownSpells = new()
		
		local i = 1
		while true do
			local name = GetSpellName(i, BOOKTYPE_SPELL)
			if not name then
				break
			end
			knownSpells[name] = true
			i = i + 1
		end
		
		for buff in pairs(self.allBuffs) do
			local found = knownSpells[buff.name]
			if found and not buff.found then
				-- SimpleSelfRebuff:Debug('New spell: %q', buff.name)
				buff.found = true
				dirty = true
			elseif not found and buff.found then
				-- SimpleSelfRebuff:Debug('Old spell: %q', buff.name)
				buff.found = nil
				dirty = true
			end
		end
		
		knownSpells = del(knownSpells)
	
		if dirty then
			SimpleSelfRebuff:AvailableBuffsChanged()
		end	
	end
	
	function SpellBuff:IsUsable()
		local start, duration = GetSpellCooldown(self.name)
		if start > 0 and duration > 0 and start + duration > GetTime() then
			return false
		end					
		return (IsUsableSpell(self.name))
	end
	
	function SpellBuff:SetupBuffSource(castButton)
		SimpleSelfRebuff:Debug('Setup casting for spell %q', self.name)
		castButton:SetAttribute('type', 'spell')
		castButton:SetAttribute('spell', self.name)
	end
	
end

-------------------------------------------------------------------------------
-- Auras
-------------------------------------------------------------------------------

loadFuncs[TARGET_AURA] = function()
	SimpleSelfRebuff:Debug('Loading aura buff support')

	local AuraBuff = BuffTargetMixin(TARGET_AURA)
	
	function AuraBuff:EnableTimedBuffWatch(flag)
		if flag then
			if not SimpleSelfRebuff:IsEventScheduled('SimpleSelfRebuff_WatchTimedBuff') then
				SimpleSelfRebuff:ScheduleRepeatingEvent('SimpleSelfRebuff_WatchTimedBuff', SimpleSelfRebuff.ScanPlayerBuffs, 1, SimpleSelfRebuff)
			end
		else
			if SimpleSelfRebuff:IsEventScheduled('SimpleSelfRebuff_WatchTimedBuff') then
				SimpleSelfRebuff:CancelScheduledEvent('SimpleSelfRebuff_WatchTimedBuff')
			end
		end
	end

	function AuraBuff:OnDisable()
		self:EnableTimedBuffWatch(false)
	end

	function AuraBuff:ScanCurrentBuffs(callback)
		local hasTimedBuff
		for i = 1, 32 do
			-- Only take in account buffs we actually cast
			local buffName, _, _, _, duration, timeLeft = UnitBuff('player', i)	
			if not buffName then
				return
			end
			local buff = buffName and self.buffs[buffName]
			if duration and buff then
				if timeLeft > 0 then
					hasTimedBuff = true
				end
				callback(buff, timeLeft)
			end
		end	
		self:EnableTimedBuffWatch(hasTimedBuff)
	end
	
	function AuraBuff:SetupBuffTarget(castButton)
		castButton:SetAttribute('unit', 'player')
	end
	
end

-------------------------------------------------------------------------------
-- Tracking
-------------------------------------------------------------------------------

loadFuncs[TARGET_TRACKING] = function()
	SimpleSelfRebuff:Debug('Loading tracking support')
	
	local TrackingBuff = BuffTargetMixin(TARGET_TRACKING)
	TrackingBuff.textures = {}
	
	function TrackingBuff:OnInstanceInit(buff)
		local texture = BS:GetSpellIcon(buff.name)
		if not texture then
			error('Unknown tracking texture for spell %q', buff.name)
		elseif self.textures[texture] then
			error('Tracking texture already set for %q: %q', buff.name, self.textures[texture].name)
		end
		self.textures[texture] = buff
	end
	
	function TrackingBuff:ScanCurrentBuffs(callback)
		local texture = GetTrackingTexture()
		local buff = texture and self.textures[texture]
		if buff then
			callback(buff)
		end
	end
	
	function TrackingBuff:SetupBuffTarget(castButton)
		castButton:SetAttribute('unit', 'player')
	end

end

-------------------------------------------------------------------------------
-- Weapon enchantment buff
-------------------------------------------------------------------------------

-- Thoses are pretty ugly to handle due to poor API
loadFuncs[TARGET_MAINHAND] = function()
	SimpleSelfRebuff:Debug('Loading weapon buff support')
	
	local weaponEnchants = {}
	
	local function ScanWeaponBuffs(self, callback)
		local timeLeft = weaponEnchants[self.target]
		if timeLeft then
			-- Assume it's the buff we want
			local anyBuff = select(2, next(self.buffs))
			local category = anyBuff.category
			local buff = category:GetExpectedBuff()
			if buff then
				callback(buff, timeLeft)
			end
		end
	end
	
	local function SetupWeaponBuff(self, castButton)
		SimpleSelfRebuff:Debug('Setting up for buff %q on %q', self.name, self.slot)
		castButton:SetAttribute('target-slot', self.slot)
	end
	
	local RefreshWeaponBuffs
	do
		local dirty
		
		local function RefreshOneWeaponBuff(target, hasEnchant, expiration)
			if hasEnchant then
				if not weaponEnchants[target] then
					dirty = true
				end
				weaponEnchants[target] = expiration/1000
			elseif not hasEnchant and weaponEnchants[target] then
				weaponEnchants[target] = nil
				dirty = true
			end
		end
		
		function RefreshWeaponBuffs()
			dirty = nil
			RefreshOneWeaponBuff(TARGET_MAINHAND, GetWeaponEnchantInfo())
			RefreshOneWeaponBuff(TARGET_OFFHAND, select(4, GetWeaponEnchantInfo()))
			if dirty then
				SimpleSelfRebuff:ScanPlayerBuffs()
			end
		end
	end
	
	local CheckWeaponSlots, IsWeaponTarget
	do
		local weaponItems = {}
		local isValidItemEquipLoc = {
			[16] = {
				INVTYPE_WEAPON         = true,
				INVTYPE_2HWEAPON       = true,
				INVTYPE_WEAPONMAINHAND = true,
			},
			[17] = {
				INVTYPE_WEAPON         = true,
				INVTYPE_WEAPONOFFHAND  = true,
			}
		}
		
		function CheckWeaponSlots()
			local dirty
			for slot=16,17 do
				local itemLink = GetInventoryItemLink("player", slot)
				if itemLink ~= weaponItems[slot] then
					weaponItems[slot] = itemLink
					dirty = true
				end
			end
			if dirty then
				SimpleSelfRebuff:AvailableBuffsChanged()
			end
		end
		
		function IsWeaponTarget(self)
			local itemLink = weaponItems[self.slot]
			local itemEquipLoc = itemLink and select(9, GetItemInfo(itemLink))
			return itemEquipLoc and isValidItemEquipLoc[self.slot][itemEquipLoc]
		end
	end
	
	local OnWeaponBuffEnable, OnWeaponBuffDisable
	do
		local registerCount = 0
	
		function OnWeaponBuffEnable(self)
			if registerCount == 0 then
				SimpleSelfRebuff:ScheduleRepeatingEvent('SimpleSelfRebuff_ScanWeaponBuffs', RefreshWeaponBuffs, 0.5)
				SimpleSelfRebuff:RegisterBucketEvent('UNIT_INVENTORY_CHANGED', 0.5, CheckWeaponSlots)
				CheckWeaponSlots()
				RefreshWeaponBuffs()
			end
			registerCount = registerCount + 1
		end
	
		function OnWeaponBuffDisable(self)
			registerCount = registerCount - 1
			if registerCount == 0 then
				SimpleSelfRebuff:CancelScheduledEvent('SimpleSelfRebuff_ScanWeaponBuffs')			
				SimpleSelfRebuff:UnregisterBucketEvent('UNIT_INVENTORY_CHANGED')
			end
		end
	end
	 
	local function WeaponBuffMixin(target, slotName)
		local slot = GetInventorySlotInfo(slotName)
		if type(slot) ~= 'number' then
			error("Invalid slot name %q => %q", slotName, slot)
		end
		local mixin = BuffTargetMixin(target, "slot")
		mixin.slot = slot
		mixin.OnEnable = OnWeaponBuffEnable
		mixin.OnDisable = OnWeaponBuffDisable		
		mixin.IsTargetCastable = IsWeaponTarget
		mixin.ScanCurrentBuffs = ScanWeaponBuffs
		mixin.SetupBuffTarget = SetupWeaponBuff
		return mixin
	end
	
	WeaponBuffMixin(TARGET_MAINHAND, 'MainHandSlot')
	WeaponBuffMixin(TARGET_OFFHAND, 'SecondaryHandSlot')
	
end

loadFuncs[TARGET_OFFHAND] = loadFuncs[TARGET_MAINHAND]

-------------------------------------------------------------------------------
-- Buff registry
-------------------------------------------------------------------------------

do
	local setupFuncs = {}

	function SimpleSelfRebuff:RegisterBuffSetup(func)
		if type(func) == "function" then
			tinsert(setupFuncs, func)
		else
			error("Arg #1 to RegisterBuffs must be a function, not %q", type(func))
		end
	end
	
	function SimpleSelfRebuff:ProcessRegistry()
		if not setupFuncs then
			return
		end
		
		-- Prebuild categories
		self:GetCategory(self.CATEGORY_TRACKING, self.TARGET_TRACKING)
		self:GetCategory(self.CATEGORY_MAINHAND, self.TARGET_MAINHAND)
		self:GetCategory(self.CATEGORY_OFFHAND,  self.TARGET_OFFHAND)
		
		-- Load all funcs
		local selfClassName = select(2, UnitClass('player'))
		for i,func in ipairs(setupFuncs) do
			local succes, msg = pcall(func, self, BS, selfClassName, L)
			if not succes then
				geterrorhandler()(("Error in setupFunc #%d: %s"):format(i, msg))
			end
		end
		
		-- Clean up
		setupFuncs = del(setupFuncs)
		loadFuncs = del(loadFuncs)
		
		self:BuildOptions()
	end
end
	
function SimpleSelfRebuff:GetCategory(name, ...)
	if type(name) ~= 'string' then
		error("Argument #2 to GetCategory should be a string, not %q", type(name))
	end
	if not categories[name] then
		categories[name] = CategoryClass:new(name, ...)
	end
	return categories[name]
end

function SimpleSelfRebuff:AddStandaloneBuff(name, ...)
	if type(name) ~= 'string' then
		error("Argument #2 to AddStandaloneBuff should be a string, not %q", type(name))
	end
	self:GetCategory(name):add(name, ...)
end

function SimpleSelfRebuff:AddMultiStandaloneBuffs(...)
	for i=1,select('#', ...) do
		local name = select(i, ...)
		self:AddStandaloneBuff(name)
	end
end

-------------------------------------------------------------------------------
-- Activation
-------------------------------------------------------------------------------

function SimpleSelfRebuff:SetMonitoringActive(value)
	if value and not monitoringActive then
		monitoringActive = true
		self:Debug('Activated')
		
		for name,source in pairs(sources) do
			if type(source.OnEnable) == "function" then
				source:OnEnable()
			end
		end		
		for name,target in pairs(targets) do
			if type(target.OnEnable) == "function" then
				target:OnEnable()
			end
		end		
	
		self:ScheduleRepeatingEvent('SimpleSelfRebuff_CheckRebuff', self.CheckRebuff, 1.5, self)
		self:ScanPlayerBuffs()
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
		
	elseif not value and monitoringActive then
		monitoringActive = nil
		self:Debug('Disactivated')

		self:CancelScheduledEvent('SimpleSelfRebuff_CheckRebuff')

		for name,source in pairs(sources) do
			if type(source.OnDisable) == "function" then
				source:OnDisable()
			end
		end		
		for name,target in pairs(targets) do
			if type(target.OnDisable) == "function" then
				target:OnDisable()
			end
		end
		
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
	end
end

function SimpleSelfRebuff:CheckActivation()
	self:SetMonitoringActive(
		not (UnitIsDeadOrGhost('player') or IsResting())
	)
end

function SimpleSelfRebuff:IsMonitoringActive()
	return monitoringActive
end

-------------------------------------------------------------------------------
-- Core
-------------------------------------------------------------------------------

function SimpleSelfRebuff:CheckRebuff()

	local rebuffThreshold = db.rebuffThreshold > 0 and db.rebuffThreshold or nil
	local newBuffToCast, buffTimeLeft, shortestBuff, shortestBuffDuration
	
	for categoryName,category in pairs(categories) do
		local expectedBuff = category:GetExpectedBuff()
		if expectedBuff and expectedBuff:IsUsable() then			
			local actualBuff, _, timeLeft = category:GetActualBuff()
			
			-- Check if it is missing
			if not actualBuff then
				newBuffToCast = expectedBuff
				--self:Debug('Missing buff: %q', expectedBuff)
				break
				
			-- Check rebuffing threshold
			elseif rebuffThreshold and timeLeft and actualBuff == expectedBuff then
				if not shortestBuffDuration or timeLeft < shortestBuffDuration then
					shortestBuffDuration, shortestBuff = timeLeft, actualBuff
				end
			end
				
		end
	end
	
	if not newBuffToCast and shortestBuff and shortestBuffDuration < rebuffThreshold then
		newBuffToCast = shortestBuff
		buffTimeLeft = shortestBuffDuration
		self:Debug('Buff needs refresh: %q (%q)', newBuffToCast, shortestBuffDuration)
	end

	if newBuffToCast ~= buffToCast then
		buffToCast = newBuffToCast
		self:TriggerEvent('SimpleSelfRebuff_PendingCastChanged', buffToCast, buffTimeLeft)
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
	end

end

function SimpleSelfRebuff:GetPendingCast()
	return buffToCast
end

-------------------------------------------------------------------------------
-- Player buff status tracking
-------------------------------------------------------------------------------

do
	local seen
	local dirty
	
	local function seenBuff(buff, timeLeft)
		--SimpleSelfRebuff:Debug('Seen %s %q (%q)', buff.target, buff.name, timeLeft)
		local category = buff.category
		
		-- Handle subcategories: if there is an expected buff and it has a
		-- subcategory, ignore the detected buff if its subcategory mismatchs
		local expected = category:GetExpectedBuff()
		if expected and expected.subcat and expected.subcat ~= buff.subcat then
			return
		end
		
		local timeLeft = timeLeft and math.floor(timeLeft) or 0
		local expiration = (timeLeft > 0) and (GetTime() + timeLeft)
		seen[category] = true
		if category:SetActualBuff(buff, expiration) then
			dirty = true
		end
	end
	
	function SimpleSelfRebuff:ScanPlayerBuffs()
		if not monitoringActive or seen then
			return
		end
		seen = new()
		dirty = nil
		
		for target, mixin in pairs(targets) do
			-- self:Debug('Scanning %s', mixin.target)
			mixin:ScanCurrentBuffs(seenBuff)
		end
		
		for categoryName, category in pairs(categories) do
			if not seen[category] then
				if category:SetActualBuff(nil, nil) then
					dirty = true
				end
			end
		end
		
		if dirty then
			dirty = nil
			self:CheckRebuff()
		end
		
		seen = del(seen)		
	end
	
end
